import numpy as np
from matplotlib import pyplot as plt
from numba import njit
import numba


@numba.vectorize(['float64(float64)'], target='cpu', nopython=True, fastmath=True)
def exp_nb(x):
	return np.exp(x)


@numba.vectorize(['float64(float64)'], target='cpu', nopython=True, fastmath=True)
def log_nb(x):
	return np.log(x)


@njit(nogil=True, cache=True, fastmath=True, looplift=True)
def exp_custom(x):
	return exp_nb(x)


@njit(nogil=True, cache=True, fastmath=True, looplift=True)
def log_custom(x):
	return log_nb(x)


@njit(nogil=True, cache=True, fastmath=True, looplift=True)
def tanh_custom(x):
	return (1 - 2 / (exp_nb(2 * x) + 1))


@njit(nogil=True, cache=True, fastmath=True)
def generate_time_series(J_ij_input, H_i_phases, number_of_phase, time_step):
	N = len(J_ij_input)

	number_of_sample = number_of_phase

	m_i = np.zeros((number_of_sample, N))
	C = np.zeros((number_of_sample, N, N))
	D = np.zeros((number_of_sample, N, N))
	S = np.zeros((number_of_sample, time_step, N))

	prob_x = np.zeros(N)

	for r in range(number_of_sample):
		for i in range(N):
			S[r, 0, i] = np.random.randint(2) * 2 - 1
		H_i_input = np.copy(H_i_phases[r])
		for t in range(time_step - 1):
			h_vec = H_i_input + np.dot(J_ij_input, S[r, t + 0, :])
			caltmp = (2 * h_vec)
			for i, v in enumerate(caltmp):
				if v > 40:
					prob_x[i] = 1
				else:
					if v < -40:
						prob_x[i] = 0
					else:
						prob_x[i] = 1 / (exp_custom(-v) + 1)
			randtmp = np.random.rand(N)
			for i in range(N):
				S[r, t + 1, i] = -1
				if randtmp[i] < prob_x[i]:
					S[r, t + 1, i] = 1
			m_i[r] += S[r, t + 1, :]
			tmpC = np.array([[(S[r, t + 1, i] * S[r, t + 1, j]) for j in range(N)] for i in range(N)])
			C[r] += tmpC
			tmpD = np.array([[(S[r, t + 1, i] * S[r, t + 0, j]) for j in range(N)] for i in range(N)])
			D[r] += tmpD

		for i in range(N):
			m_i[r, i] = m_i[r, i] / (time_step - 1)
		for i in range(N):
			for j in range(N):
				C[r, i, j] = C[r, i, j] / ((time_step - 1)) - m_i[r, i] * m_i[r, j]
				D[r, i, j] = D[r, i, j] / ((time_step - 1)) - m_i[r, i] * m_i[r, j]

	return m_i, C, D, S


def turn_activities_into_matrix(activities, bin_size, max_n=None, max_t=None):
	if max_t is None:
		t_size = activities[:, 1].max() + bin_size
	else:
		t_size = max_t + bin_size
	bins = np.arange(0, t_size, bin_size)
	time_instance = np.digitize(activities[:, 1], bins) - 1
	neurons = (activities[:, 0] - activities[:, 0].min()).astype(int)
	if max_n is None:
		n_size = neurons.max() + 1
	else:
		n_size = max_n
	S = -1 * np.ones((len(bins), n_size), dtype=np.float64)
	for i, n in enumerate(neurons):
		S[time_instance[i], n] = 1

	total_spikes = np.sum((S + 1) // 2, axis=1)
	sparsity = np.sum(total_spikes == 0) / len(total_spikes)

	boolean_array = (total_spikes == 0)
	result_array = boolean_array[1:] & boolean_array[:-1]
	result_array = np.insert(result_array, 0, False)
	new_S = S[np.invert(result_array), :]

	new_total_spikes = np.sum((new_S + 1) // 2, axis=1)
	new_sparsity = np.sum(new_total_spikes == 0) / len(new_total_spikes)

	# return S , sparsity, new_S, new_sparsity
	return new_S, sparsity


sim_time = 3600
step_size = 0.005

fname = 'control_0'

spikes_fname = 'control_data_0_spikes.csv'
monte_spikes_all_fname = 'control_data_0_monte_spikes'

for methods in ['MLE', 'MAP', 'ML-II']:

	J_fname = '{}_J_est_{}.txt'.format(fname, methods)
	H_fname = '{}_H_est_{}.txt'.format(fname, methods)

	monte_spikes_fname = monte_spikes_all_fname + '_' + methods + '.csv'

	sim_steps = int(sim_time / step_size)

	J_ij = np.loadtxt(J_fname)
	H_i = np.loadtxt(H_fname)
	H_i = np.array([H_i])

	m_i, C, D, S = generate_time_series(J_ij, H_i, 1, sim_steps)
	S = (S + 1) // 2

	monte_spikes = np.r_[[np.where(S[0])[0], np.where(S[0])[1]]].astype(float).T
	# monte_spikes[:,0] = monte_spikes[:,0]*step_size

	true_spikes = np.loadtxt(spikes_fname, delimiter=',')
	true_S, sparse_P = turn_activities_into_matrix(true_spikes, step_size, max_n=None, max_t=None)
	true_S = (true_S + 1) // 2

	padding = np.random.geometric(p=1 - sparse_P, size=len(S[0]))
	padding = padding * step_size
	for i in range(1, len(padding)):
		padding[i] = step_size + padding[i - 1] + padding[i]
	monte_spikes[:, 0] = padding[monte_spikes[:, 0].astype(int)]

	monte_spikes = monte_spikes[monte_spikes.T[0] < 3600, :]
	np.savetxt(monte_spikes_fname, monte_spikes[:, [1, 0]], delimiter=',', header='id, time')

import plotly.graph_objs as go
import plotly.io as pio

# Sample data
true_spikes = true_spikes[:, [1, 0]]
true_spikes = true_spikes[true_spikes.T[0] < 1000, :]
x_data2 = true_spikes.T[0]
y_data2 = true_spikes.T[1]

# Create traces
trace2 = go.Scatter(
	x=x_data2,
	y=y_data2,
	mode='markers',
	name='true'
)


trace = [trace2]
for methods in ['MLE', 'MAP', 'ML-II']:
	monte_spikes_fname = monte_spikes_all_fname + '_' + methods + '.csv'
	monte_spikes = np.loadtxt(monte_spikes_fname, delimiter=',', skiprows=1)
	monte_spikes = monte_spikes[:, [1, 0]]

	monte_spikes = monte_spikes[monte_spikes.T[0] < 1000, :]
	x_data1 = monte_spikes.T[0]
	y_data1 = monte_spikes.T[1]
	trace1 = go.Scatter(
		x=x_data1,
		y=y_data1,
		mode='markers',
		name=methods
	)
	trace.append(trace1)

# Create layout
layout = go.Layout(
	title='Raster plot',
	xaxis=dict(title='time (s)'),
	yaxis=dict(title='ROI id')
)

# Create figure
fig = go.Figure(data=trace, layout=layout)

# Show plot
pio.write_html(fig, 'scatter_plot.html')
